/* $Id: load.c,v 1.5 1999/08/02 17:40:45 stuart Exp $
 *
 * sh3mon/load.c
 *
 * Copyright (C) 1999 Stuart Menefy
 */

#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>
#include "sh3mon.h"

#define TARGET "sh-unknown-elf"

typedef struct {
  void (*fn)(const unsigned char* buffer, int len, unsigned long addr);
} iterator_data_t;

static bfd* abfd;

static void fatal (const char *string);
static void load_iterator (bfd *abfd, asection *section, PTR data);

void
load_open(const char *filename)
{
  bfd_init ();

  if (! bfd_set_default_target (TARGET))
    {
      char errmsg[100];
      sprintf (errmsg, "can't set BFD default target to `%s'", TARGET);
      fatal (errmsg);
    }

  abfd = bfd_openr (filename, NULL);
  if (abfd == NULL)
    {
      fatal (filename);
    }
}

unsigned long
load_write(void (*fn)(const unsigned char* buffer, int len, unsigned long addr))
{
  bfd_vma entry;
  iterator_data_t data;

  data.fn = fn;

  if (! bfd_check_format (abfd, bfd_object))
    {
      fatal (bfd_get_filename (abfd));
    }

  bfd_map_over_sections (abfd, load_iterator, &data);

  entry = bfd_get_start_address(abfd);
  return entry;
}

void
load_close(void)
{
  if (bfd_close (abfd) == false)
    {
      fatal ("filename");
    }
}

static void
load_iterator (
     bfd *abfd,
     asection *section,
     PTR ptr)
{
  flagword flags;
  bfd_size_type datasize;
  bfd_byte *data;
  iterator_data_t* iterator_data = (iterator_data_t*)ptr;
  const char* load_sections[] = {
    ".text", ".rodata", "__ex_table", ".data", ".data.init_task", ".text.init", ".data.init"
  };
  int a;
  int load = 0;

  flags = bfd_get_section_flags (abfd, section);

  datasize = bfd_get_section_size_before_reloc (section);

  printf("Section: %s (at 0x%08lx) 0x%08lx bytes\n",
	 section->name, section->vma, datasize);

  if (datasize == 0) {
    return;
  }

  for (a=0; a<(sizeof(load_sections) / sizeof(char*)); a++) {
    if (strcmp(section->name, load_sections[a]) == 0) {
      load = 1;
      break;
    }
  }

  if (!load)
    return;

  data = (bfd_byte *) malloc ((size_t) datasize);
  if (data == NULL) {
    fprintf(stderr, "%s: Out of memory\n", program_name);
    exit(1);
  }
  bfd_get_section_contents (abfd, section, data, 0, datasize);

  iterator_data->fn(data, datasize, section->vma);
}

static void
fatal (const char *string)
{
  CONST char *errmsg = bfd_errmsg (bfd_get_error ());

  if (string)
    fprintf (stderr, "%s: %s: %s\n", program_name, string, errmsg);
  else
    fprintf (stderr, "%s: %s\n", program_name, errmsg);
  exit(1);
}
